/*----------------------------------------------------------------------------
    calories - real-time, networked, cellular automaton simulation

    Copyright (c) 2008 Ge Wang
        All rights reserved.
        http://ccrma.stanford.edu/~ge/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: ca.h
// desc: header file for cellular automaton classes
//
// author: Ge Wang (ge@ccrma.stanford.edu)
//   date: 2008.1.15
//-----------------------------------------------------------------------------
#ifndef __CA_H__
#define __CA_H__


// includes
#include "calories_def.h"
#include "vectors.h"


// directions
#define LEFT -1
#define RIGHT 1
#define BELOW -1
#define ABOVE 1
#define AWAY -1
#define TOWARD 1




//-----------------------------------------------------------------------------
// name: struct Cell
// desc: a cell in CA
//-----------------------------------------------------------------------------
struct Cell
{
    long value;     // integer value
    double valuef;  // floating point value
    float size;     // current visual size
    float sizeto;   // current visual target to interpolate towards
    long x;         // x coordinate in CA
    long y;         // y coordinate in CA
    long z;         // z coordinate in CA

    // constructor
    Cell()
    : value(0), valuef(0), size(0), sizeto(0), x(0), y(0), z(0)
    { }
};




//-----------------------------------------------------------------------------
// name: struct CA
// desc: basic cellular automaton class
//-----------------------------------------------------------------------------
struct CA
{
    // the grid
    Cell *** grid;
    // next grid
    Cell *** next;
    // dims
    long width;
    long height;
    long depth;
    // generations
    long generations;
    // interpolation
    float slew;
    // number of currently active cells
    long numActive;

    // constructor
    CA()
    {
        // zero
        grid = NULL;
        next = NULL;
        width = height = depth = 0;
        generations = 0;
        slew = .2;  // change to adjust interpoloation speed
        numActive = 0;
    }

    // destructor
    ~CA()
    {
        // clean up
        this->cleanup();
    }

    // clean up
    void cleanup()
    {
        int x, y;

        // reclaim memory
        if( grid && next )
        {
            for( x = 0; x < width; x++ )
            {
                for( y = 0; y < height; y++ )
                {
                    SAFE_DELETE_ARRAY( grid[x][y] );
                    SAFE_DELETE_ARRAY( next[x][y] );
                }
                SAFE_DELETE_ARRAY( grid[x] );
                SAFE_DELETE_ARRAY( next[x] );
            }
            SAFE_DELETE_ARRAY( grid );
            SAFE_DELETE_ARRAY( next );
        }

        // reset
        width = height = depth = 0;
        generations = 0;
    }

    // init
    bool init( long w, long h, long d )
    {
        int x, y, z = 0;
        // clean up first
        cleanup();

        // try to allocate
        grid = new Cell**[w];
        next = new Cell**[w];
        assert( grid && next );
        // for each
        for( x = 0; x < w; x++ )
        {
            // allocate
            grid[x] = new Cell*[h];
            next[x] = new Cell*[h];
            assert( grid[x] && next[x] );
            // for each
            for( y = 0; y < h; y++ )
            {
                // allocate
                grid[x][y] = new Cell[d];
                next[x][y] = new Cell[d];
                assert( grid[x][y] && next[x][y] );
                for( z = 0; z < d; z++ )
                {
                    grid[x][y][z].x = next[x][y][z].x = x;
                    grid[x][y][z].y = next[x][y][z].y = y;
                    grid[x][y][z].z = next[x][y][z].z = z;
                }
            }
        }

        // remember
        width = w;
        height = h;
        depth = d;

        return true;
    }

    // copy
    void copy( Cell *** to, Cell *** from )
    {
        long x, y, z;

        // go
        for( x = 0; x < width; x++ )
        {
            // go
            for( y = 0; y < height; y++ )
            {
                // go
                for( z = 0; z < depth; z++ )
                {
                    to[x][y][z].value = from[x][y][z].value;
                    to[x][y][z].valuef = from[x][y][z].valuef;
                    to[x][y][z].size = from[x][y][z].size;
                    to[x][y][z].sizeto = from[x][y][z].sizeto;
                }
            }
        }
    }

    // update
    virtual void update();
    // getCell
    virtual const Cell * getCell( const Cell * from, long dx, long dy, long dz );
    // getCell
    virtual const Cell * getCell( long x, long y, long z );
    // setCell
    virtual bool setCell( long x, long y, long z, long value );
    // getNumNeighbors
    virtual long getNumNeighbors( const Cell * from );
    // pure virtual
    virtual void update( const Cell * current, Cell * next ) = 0;
    // randomize
    virtual void randomize( double prob );
    // update look
    virtual void updateLook();
};




#endif
